/** Word-based compression/decompression classes
 * 
 * <p>Classes in this package provide interfaces for the compression system, and implementations
 * for codeword-based compression. Their main usage is the construction of 
 * {@linkplain it.unimi.dsi.util.ImmutableExternalPrefixMap prefix maps}, but
 * they are also used, for instance, for <a href="http://webgraph.di.unimi.it/">WebGraph</a> label
 * compression.
 */

package it.unimi.dsi.compression;
